"""
(c) 2024 LLC «DACOM M»
"""
import dataclasses
import glob
import json
import logging
# pylint: disable=redefined-builtin
from concurrent.futures import ThreadPoolExecutor, TimeoutError
from dataclasses import dataclass, field
from itertools import chain
from multiprocessing.pool import ThreadPool
from subprocess import check_call, CalledProcessError, DEVNULL, check_output
from typing import Optional, List, Union

import os
import re
from abc import ABC, abstractmethod
from deepdiff import DeepDiff
from jinja2 import Template

from veilrpc.utils.bazalt import (
    get_result, get_output, states_result, OvsPriority, OvsTableNew, OVS_FLOW_PRESETS_NEW, OVS_TABLE_MAX_OFFSET,
    CommandExecutionError
)

NET_PARAM_BRIDGE_FLOW_FILE_PATTERN = re.compile(r'([a-f0-9]{2}){4}.json|vnet_([a-f0-9]{2}){2}(_vm){0,1}.json',
                                                re.IGNORECASE)

LOG = logging.getLogger("bazalt")
BASE_PATH = "/var/lib/lxc"
NET_PARAMS_DIR = '/var/lib/ecp-veil/node/bin/networks'
SRIOV_STATE_PARAMS = os.path.join(NET_PARAMS_DIR, 'sriov-state.json')
TRAFFIC_BLOCK_PARAMS = os.path.join(NET_PARAMS_DIR, 'traffic-block.json')
PORT_MIRROR_PARAMS = os.path.join(NET_PARAMS_DIR, 'port-mirrors.json')
TRAFFIC_SHAPING_PARAMS = os.path.join(NET_PARAMS_DIR, 'traffic-shaping.json')
VM_QOS_POL_MARK_PARAMS = os.path.join(NET_PARAMS_DIR, 'vmachines-qos-policy-marking.json')
VNETWORK_QOS_SHAPING_PARAMS = os.path.join(NET_PARAMS_DIR, 'vnetwork-qos-shaping.json')
VNETWORK_ACL_PARAMS = os.path.join(NET_PARAMS_DIR, 'vnetwork-acl.json')
SYS_CLASS_NET_DIR = '/sys/class/net'
CHECK_FIREWALL_FLAG = os.path.join(NET_PARAMS_DIR, 'check_firewall')

BASE_LXC_CONFIG = [
    'lxc.include = /usr/share/lxc/config/debian.common.conf',
    'lxc.tty.max = 4',
    'lxc.apparmor.profile = unconfined',
    'lxc.cgroup.devices.allow = a',
    'lxc.log.level = 1',
    'lxc.cgroup.memory.limit_in_bytes = 256M'
]

OVF_RULE_PATTERN = re.compile(r'.*(in_port=(?P<port>.*))(\s)(actions=(?P<action>.*)).*')
OVF_RULE_PATTERN_BLOCK = re.compile(r'.*(priority=(?P<priority>\d+))[,\s]*'
                                    r'(in_port=(?P<port>.*))(\s)(actions=(?P<action>.*)).*')
OVF_RULE_PATTERN_VM_PORT = re.compile(r'.*(in_port=(?P<in_port>\w+)|reg0=(?P<out_port>\w+)).*')
OVF_RULE_PATTERN_CONNTRACK = re.compile(r"ct_state=(?P<ct_states>.+?),")
DEFAULT_SWITCH = 'veil-default'
BLACKHOLE_SWITCH = 'veil-blackhole'


class NetworkConstants:
    """
    Базовый класс, определяющий поведение сетевой константы.
    """

    FILE_PATH = None
    CONSTANTS = None
    DEFAULT = None

    def __init__(self, data: dict = None, *args, **kwargs):
        if data is None:
            try:
                self.load()
            except IOError:
                print(f'Can\'t open {self.FILE_PATH} file')
        else:
            self.CONSTANTS = data

    def load(self):
        raise NotImplementedError

    def get(self, *args, **kwargs):
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        return {
            key: obj() if callable(obj) else obj for key, obj in self.CONSTANTS.items()
        }

    def __getattr__(self, item: str, *args, **kwargs):
        return self.get(item.lower(), *args, **kwargs)


class InternetProtocols(NetworkConstants):
    """
    Internet (IP) protocols
    Source: https://www.iana.org/assignments/protocol-numbers/protocol-numbers.xhtml
    Loaded from /etc/protocols

    Usage example:
        >>> INTERNET_PROTOCOLS = InternetProtocols()
        >>> INTERNET_PROTOCOLS()
        -> return dict with all Internet Protocols.
        >>> INTERNET_PROTOCOLS.tcp
        6
        >>> INTERNET_PROTOCOLS.get('udp')
        17
    """

    FILE_PATH = '/etc/protocols'
    CONSTANTS: dict[str, int] = dict()

    def load(self):
        with open(self.FILE_PATH) as file:
            for line in file:
                try:
                    if line[0] in ["#", "\n"]:
                        continue
                    lt = tuple(re.split("[ \t]+|\n", line))
                    if len(lt) < 3:
                        continue
                    self.CONSTANTS.update({lt[2].lower(): int(lt[1])})
                except Exception:
                    LOG.info(f'Couldn\'t parse one line from protocols file ({line!r})')

    def get(self, value: str, *args, **kwargs) -> int:
        return self.getprotobyname(value, *args, **kwargs)

    def getprotobyname(self, name: str, *args, **kwargs) -> int:
        return self.CONSTANTS.get(name, self.DEFAULT)


class EthernetTypes(NetworkConstants):
    """
    Ethernet frame types
    Source: https://www.iana.org/assignments/ieee-802-numbers/ieee-802-numbers.xhtml
    Loaded from /etc/ethertypes

    Usage example:
        >>> ETHERNET_TYPES = EthernetTypes()
        >>> ETHERNET_TYPES()
        -> return dict with all Ethernet frame types.
        >>> ETHERNET_TYPES.ipv4
        0x0800
        >>> ETHERNET_TYPES.get('arp')
        0x0806
    """

    FILE_PATH = '/etc/ethertypes'
    CONSTANTS: dict[str, hex] = dict()

    def load(self):
        with open(self.FILE_PATH) as file:
            for line in file:
                try:
                    if line[0] in ["#", "\n"]:
                        continue
                    lt = tuple(re.split("[ \t]+|\n", line))
                    if len(lt) < 2:
                        continue
                    self.CONSTANTS.update({lt[0].lower(): hex(int(lt[1], 16))})
                except Exception:
                    LOG.info(f'Couldn\'t parse one line from ethertypes file ({line!r})')

    def get(self, value: str, *args, **kwargs) -> hex:
        return self.gettypebyname(value, *args, **kwargs)

    def gettypebyname(self, name: str, *args, **kwargs) -> hex:
        return self.CONSTANTS.get(name, self.DEFAULT)


class NetworkServices(NetworkConstants):
    """
    Network services, Internet style
    Source: https://www.iana.org/assignments/service-names-port-numbers/service-names-port-numbers.xhtml
    Loaded from /etc/services

    Usage example:
        >>> NETWORK_SERVICES = NetworkServices()
        >>> NETWORK_SERVICES()
        -> return dict with all Network services.
        >>> NETWORK_SERVICES.tcp()
        -> return dict with all TCP services.
        >>> NETWORK_SERVICES.tcp.ssh
        22
        >>> NETWORK_SERVICES.udp.get('domain')
        53
        >>> NETWORK_SERVICES.get('tcp').get('git')
        9418
    """

    FILE_PATH = '/etc/services'
    PROTOS = ('tcp', 'udp', 'sctp')
    CONSTANTS: dict[str, 'NetworkServices'] = {key: dict() for key in PROTOS}

    def load(self):
        with open(self.FILE_PATH) as file:
            for line in file:
                try:
                    if line[0] in ["#", "\n"]:
                        continue
                    lt = tuple(re.split("[ \t]+|\n", line))
                    if len(lt) < 2:
                        continue
                    for proto in self.PROTOS:
                        if lt[1].endswith(f'/{proto}'):
                            self.CONSTANTS[proto].update({lt[0]: int(lt[1].split('/')[0])})
                            break
                except Exception:
                    print(f'Couldn\'t parse one line from services file ({line!r})')

        for proto in self.PROTOS:
            self.CONSTANTS[proto] = NetworkServices(data=getattr(self, proto))

    def get(self, value: str, default=None, *args, **kwargs) -> dict:
        return self.getservicebyname(value, default, *args, **kwargs)

    def getservicebyname(self, name: str, default=None, *args, **kwargs) -> dict:
        if default is None:
            default = self.DEFAULT
        return self.CONSTANTS.get(name, default)


class IcmpTypes(NetworkConstants):
    """
    Internet Control Message Protocol (ICMP) Types
    Source: https://www.iana.org/assignments/icmp-parameters/icmp-parameters.xhtml
    Static

    Usage example:
        >>> ICMP_TYPES = IcmpTypes()
        >>> ICMP_TYPES.echoreply
        0
        >>> ICMP_TYPES.get('echo')
        8
    """

    CONSTANTS: dict[str, int] = dict()

    def load(self):
        self.CONSTANTS = {
            'echoreply': 0,         # Echo Reply,[RFC792]
            'unreach': 3,           # Destination Unreachable,[RFC792]
            'sourcequench': 4,      # Source Quench (Deprecated),[RFC792][RFC6633]
            'redirect': 5,          # Redirect,[RFC792]
            'echo': 8,              # Echo,[RFC792]
            'routeradvert': 9,      # Router Advertisement,[RFC1256]
            'routersolicit': 10,    # Router Solicitation,[RFC1256]
            'timxceed': 11,         # Time Exceeded,[RFC792]
            'paramprob': 12,        # Parameter Problem,[RFC792]
            'tstamp': 13,           # Timestamp,[RFC792]
            'tstampreply': 14,      # Timestamp Reply,[RFC792]
            'ireq': 15,             # Information Request (Deprecated),[RFC792][RFC6918]
            'ireqreply': 16,        # Information Reply (Deprecated),[RFC792][RFC6918]
            'maskreq': 17,          # Address Mask Request (Deprecated),[RFC950][RFC6918]
            'maskreply': 18         # Address Mask Reply (Deprecated),[RFC950][RFC6918]
        }

    def get(self, value: str, *args, **kwargs) -> int:
        return self.gettypebyname(value, *args, **kwargs)

    def gettypebyname(self, name: str, *args, **kwargs):
        return self.CONSTANTS.get(name, self.DEFAULT)


def interface_exists(name: str, mac_address: str, *args, **kwargs) -> bool:
    """
    Функция осуществляет проверку существования интерфейса на узле по его наименованию и MAC адресу.
    """

    return (get_result(f'ip -br link | egrep "{name}.*{mac_address}"') or
            get_result(f'ovs-vsctl get Interface {name} external-ids:attached-mac | grep {mac_address}'))


# TODO такое преобразование понадобится, если в утилите tc задавать vlan, но учитывать 16bit поле
#  (ex: tc filter add... protocol 802.1q u32 match u16 0x0064 0x0fff at -4), а не 12 bit
def convert_dec_12bit_to_hex_16bit(dec):
    """
    Функция преобразует десятичное значение для 12 bit в hex значение 16 bit
    :param dec: integer
    :return: hex
    """

    if dec not in range(0, 4096):
        raise Exception('The decimal integer should be in range between 0 to 4095')
    return f"{dec:#0{6}x}"


def convert_dscp_dec_to_tos_hex(dec):
    """
    Функция преобразует десятичное значение для 6-ти битного подполя DSCP ipv4 заголовка
    в hex значение 8 битного поля ToS
    :param dec: decimal integer
    :return: hex
    """

    if dec not in range(0, 64):
        raise Exception('The decimal integer should be in range between 0 to 63')
    return f"{dec << 2:#0{4}x}"


def create_qos_marking_rule_to_valid_str(rule: dict, vminf: str) -> str:
    rule_struct = {'direction': rule['direction'],
                   'vminf': vminf,
                   'pref_index': rule['pref_index'],
                   'dscp': rule.get('dscp'),
                   'cos': rule.get('cos'),
                   'ethertype': rule['ethertype'],
                   'vlan_id': rule.get('vlan_id'),
                   'src_mac': rule.get('src_mac'),
                   'dst_mac': rule.get('dst_mac'),
                   'ip_protocol': rule.get('ip_protocol'),
                   'src_ip': rule.get('src_ip'),
                   'dst_ip': rule.get('dst_ip'),
                   'src_port': rule.get('src_port'),
                   'dst_port': rule.get('dst_port'),
                   }

    if rule_struct['dscp']:
        rule_struct['dscp'] = convert_dscp_dec_to_tos_hex(rule_struct['dscp'])
    elif rule_struct['vlan_id']:
        rule_struct['vlan_id'] = convert_dec_12bit_to_hex_16bit(rule_struct['vlan_id'])

    temp = Template('{% if direction == "ingress" %}'
                    'tc filter add dev {{vminf}} parent ffff: pref {{pref_index}} '
                    # 'tc filter replace dev {{vminf}} parent ffff: pref {{pref_index}} '
                    '{% elif direction == "egress" %}'
                    'tc filter add dev {{vminf}} parent 1: pref {{pref_index}} '
                    # 'tc filter replace dev {{vminf}} parent 1: pref {{pref_index}} '
                    '{% endif %}'
                    '{% if ethertype == "ip" %}'
                    'protocol ip u32 '
                    '{% if ip_protocol %}'
                    '{% if ip_protocol == "tcp"%}'
                    'match ip protocol 6 0xff '
                    '{% elif ip_protocol == "udp"%}'
                    'match ip protocol 17 0xff '
                    '{% elif ip_protocol == "icmp"%}'
                    'match ip protocol 1 0xff '
                    '{% endif %}'
                    '{% endif %}'
                    '{% if dst_ip %}'
                    'match ip dst {{dst_ip}} '
                    '{% endif %}'
                    '{% if src_ip %}'
                    'match ip src {{src_ip}} '
                    '{% endif %}'
                    '{% if dst_port %}'
                    'match ip dport {{dst_port}} 0xffff '
                    '{% endif %}'
                    '{% if src_port %}'
                    'match ip sport {{src_port}} 0xffff '
                    '{% endif %}'
                    '{% if dst_mac %}'
                    'match ether dst {{dst_mac}} '
                    '{% endif %}'
                    '{% if src_mac %}'
                    'match ether src {{src_mac}} '
                    '{% endif %}'
                    '{% if not ip_protocol and not src_ip and not dst_ip '
                    'and not src_port and not dst_port and not dst_mac and not src_mac %}'
                    'match u32 0 0 '
                    '{% endif %}'
                    '{% endif %}'
                    '{% if dscp %}'
                    'action pedit munge ip tos set {{dscp}} retain 0xfc pipe csum ip'
                    '{% endif %}'
                    )
    rule_str = temp.render(rule_struct)
    return rule_str


def get_openflow_priority_with_offset(rule_index: int) -> tuple[int, int]:
    """
    Функция возвращает приоритет и смещение в таблице для
    OpenFlow правила на основе пользовательского индекса.
    """
    priority = (max(OvsPriority.user_acl) - rule_index) & OvsPriority.pmask
    offset = (OVS_TABLE_MAX_OFFSET - (((max(OvsPriority.user_acl) - rule_index) & OvsPriority.omask) >> 16))
    return priority, offset


def flow_in_set(flow, flows):
    """
    Функция проверки наличия правила в наборе правил.
    Позволяет выполнять поиск с учетом произвольного порядка условий.
    """
    flows = list(flows)
    flow = diff_flows_helper(flow)
    for i in range(len(flows)):
        if flow == diff_flows_helper(flows[i].replace("delete_strict ", "").replace("add ", "")):
            return True
    return False


def diff_flows_helper(flow):
    """
    Вспомогательная функция для преобразования правила во множество
    """
    flow = flow.replace(", ", ",").replace(" ", ",")
    ct_states = OVF_RULE_PATTERN_CONNTRACK.search(flow)
    if ct_states:
        flow = flow.replace(f"ct_state={ct_states.group('ct_states')},", "")
        ct_states = tuple(sorted(ct_states.group('ct_states').split("+")[1:]))
    else:
        ct_states = tuple()
    flow = tuple(flow.split(","))
    return tuple(sorted(flow)) + ct_states


def diff_flows(flows_a, flows_b):
    """
    Функция выполняет diff между двумя наборами правил.
    Позволяет выполнять diff с учетом произвольного порядка условий.
    """
    result = []
    tmp_a = {}
    tmp_b = {}
    len_a = len(flows_a)
    len_b = len(flows_b)
    count = len_a if len_a > len_b else len_b
    flows_a = list(flows_a)
    flows_b = list(flows_b)
    for i in range(count):
        if i < len_a:
            tmp_a[diff_flows_helper(flows_a[i])] = flows_a[i]
        if i < len_b:
            tmp_b[diff_flows_helper(flows_b[i])] = flows_b[i]
    diff_keys = tmp_a.keys() - tmp_b.keys()
    for item in diff_keys:
        result.append(tmp_a[item])
    del tmp_a, tmp_b
    return set(result)


def make_flows_uniq(flows):
    """
    Функция возвращает список уникальных OpenFlow правил.
    Позволяет искать совпадающие правила с учетом произвольного порядка условий.
    """
    pattern_tb = re.compile(r"table=(?P<table>\d+)")
    result = []
    tmp = []
    for flow in flows:
        if "delete_strict " in flow[:14]:
            result.append(flow)
        elif "add " in flow[:4]:
            flow = flow.replace("add ", "").replace(", ", ",").replace(" ", ",")
            ct_states = OVF_RULE_PATTERN_CONNTRACK.search(flow)
            flow_no_ct = flow
            if ct_states:
                flow_no_ct = flow_no_ct.replace(f"ct_state={ct_states.group('ct_states')},", "")
                ct_states = tuple(sorted(ct_states.group('ct_states').split("+")[1:]))
            else:
                ct_states = tuple()
            flow_no_ct = tuple(flow_no_ct.split(","))
            sorted_flow = tuple(sorted(flow_no_ct)) + ct_states
            if sorted_flow not in tmp:
                tmp.append(sorted_flow)
                match = pattern_tb.search(flow).group('table')
                flow = flow.replace(f"table={match},", f"table={match}, ").replace(",actions=", " actions=")
                result.append(f"add {flow}")
    return result


def convert_vm_out_flow(flow):
    """
    Функция конвертирует OpenFlow правило, заменяя значение регистра 'reg0'
    на число, которое соответствует имени интерфейса в OvS.
    """
    pattern_port = re.compile(r"reg0=(?P<out_port>\w+)")
    match = pattern_port.search(flow)
    if out_port := match.group('out_port'):
        response = states_result(get_output, cmd=f'ovs-vsctl --if-exists get Interface {out_port} ofport').splitlines()
        if response:
            flow = flow.replace(f"reg0={out_port}", f"reg0={hex(int(response[0]))}")
    return flow


def get_interface_name_by_number(bridge: str, number: int):
    """
    Функция возвращает имя интерфейса в OvS по его номеру
    """

    # ToDo: returns a list of interfaces with the same ofport value. Need to find a better solution.
    # response = states_result(get_output,
    #                          cmd=f'ovs-vsctl --no-head --columns=name find Interface ofport={number}').splitlines()
    # if response:
    #     return response[0]

    response = states_result(get_output,
                             cmd=f"ovs-ofctl dump-ports-desc {bridge} | grep -v LOCAL | grep addr").splitlines()
    if response:
        for i, _ in enumerate(response):
            port = re.split(r":", re.sub(r"[()]", ":", response[i].strip()))
            try:
                if int(port[0]) == number:
                    return port[1]
            except ValueError:
                pass
    return None


def delete_qos_marking_rule_to_valid_str(rule: dict, vminf: str) -> str:
    rule_struct = {'direction': rule['direction'],
                   'vminf': vminf,
                   'pref_index': rule['pref_index'],
                   'ethertype': rule['ethertype'],
                   }

    temp = Template('{% if direction == "ingress" %}'
                    'tc filter delete dev {{vminf}} parent ffff: '
                    '{% elif direction == "egress" %}'
                    'tc filter delete dev {{vminf}} parent 1: '
                    '{% endif %}'
                    '{% if ethertype == "ip" %}'
                    'protocol ip '
                    '{% endif %}'
                    'pref {{pref_index}} u32'
                    )
    rule_str = temp.render(rule_struct)
    return rule_str


def get_port_qos(port: str) -> dict:
    """
    Функция возвращает конфигурацию qos для порта
    """
    cmd = f"ovs-vsctl -f json list {{type}} {port} | jq"
    result = dict()
    try:
        port_info = json.loads(states_result(get_output, cmd=cmd.format(type='port')))
        inf_info = json.loads(states_result(get_output, cmd=cmd.format(type='interface')))
    except json.decoder.JSONDecodeError:
        result['qos'] = None
        result['rate'] = 0
    else:
        idx_egress = port_info['headings'].index('qos')
        idx_ingress = inf_info['headings'].index('ingress_policing_rate')
        egress = port_info['data'][0][idx_egress][1]
        result['qos'] = egress if not isinstance(egress, list) else None
        result['rate'] = int(inf_info['data'][0][idx_ingress]) * 1000
    return result


def get_vnetworks() -> list:
    """
    Функция возвращает список коммутаторов виртуальных сетей
    """
    response = states_result(get_output, cmd='ovs-vsctl list-br').splitlines()
    is_valid = lambda x: x.startswith('vnet_') or (x is not DEFAULT_SWITCH and x is not BLACKHOLE_SWITCH)
    return [bridge for bridge in response if is_valid(bridge)]


def get_ports_vm(switch: str) -> list:
    """
    Функция возвращает список портов подключения ВМ к коммутатору
    """
    response = states_result(get_output, cmd=f'ovs-vsctl list-ports {switch}').splitlines()
    return list(filter(lambda x: x.startswith('vm_'), response))


def get_flows(switch: str) -> list:
    """
    Функция возвращает список правил OpenFlow коммутатора
    """
    response = states_result(get_output, cmd=f"ovs-ofctl --no-stats --names dump-flows {switch}").splitlines()
    regex = re.compile(r".*hard_timeout=\d*,\sdl_dst=([0-9a-f]{2}:){5}[0-9a-f]{2}\sactions=load.*")
    flows = filter(lambda x: not regex.match(x), response)
    result = list()
    for flow in flows:
        flow = flow.strip()
        if flow.startswith('priority'):
            flow = f"table=0,{flow}"
        result.append(flow.replace(' actions', ',actions').replace(' ', ''))
    return result


class VNService:
    """
    Класс описывающий контейнер LXC
    """

    __slots__ = ('name', 'config_file', 'path', 'config', 'status', 'networks')

    def __init__(self, name: str, initial: bool = False):
        self.name = name
        self.path = os.path.join(BASE_PATH, name)
        self.config_file = os.path.join(self.path, 'config')
        self.config = self._get_config(initial=initial)
        self.status = self.get_status(self.name)
        self.networks = self.get_networks()

    class STATUS:
        running = 'RUNNING'
        stopped = 'STOPPED'

    @classmethod
    def list_containers(cls, as_objects: bool = False) -> list:
        """
        Получение списка контейнеров
        :param as_objects: флаг определяющий, возвращать список контейнеров в виде списка имен или в виде списка
        экземпляров класса
        :return: возвращает список контейнеров
        """

        cmd = "lxc-ls"
        output = cls._get_shell_output(cmd)
        result = output.split() if not as_objects else [VNService(item) for item in output.split()]
        return result

    @staticmethod
    def _run_shell_cmd(cmd: str) -> bool:
        """
        Выполнение команды shell

        :param cmd: текст команды
        :return: возвращает True, если команда выполнена, и False, если выполнение команды завершилось ошибкой
        """

        try:
            check_call(cmd, shell=True, stdout=DEVNULL, stderr=DEVNULL)
            return True
        except CalledProcessError:
            return False

    @staticmethod
    def _get_shell_output(cmd: str, timeout: int = 30, universal_newlines: bool = True,
                          return_errcode: bool = False) -> str:
        """
        Выполнение команды shell и возврат вывода
        :param cmd: текст команды
        :param timeout: таймаут выполнения
        :param universal_newlines:
        :param return_errcode: флаг, определяющий возвращать текст ошибки или код возврата
        :return: возвращает вывод команды, текст ошибки или код возврата команды
        """

        try:
            output = check_output(cmd, shell=True, executable='/bin/bash', timeout=timeout,
                                  universal_newlines=universal_newlines)
            return output
        except CalledProcessError as e:
            return e.output if not return_errcode else str(e.returncode)

    def _get_config(self, initial: bool = False) -> Optional[dict]:
        """
        Преобразование текстового файла конфигурации контейнера в словарь

        :return: возвращает словарь, содержащий параметра конфигурации в качестве ключей и их значения или None, если в
        процессе чтения конфигурации возникла ошибка
        """

        if initial:
            raw_config = BASE_LXC_CONFIG
        else:
            try:
                with open(self.config_file, 'r') as f:
                    raw_config = f.read().splitlines()
            except IOError:
                return None

        config = {item.split(' = ')[0]: item.split(' = ')[1] for item in raw_config if item}
        return config

    def get_leases(self) -> list:
        headers = ('address', 'hwaddr', 'client_id', 'valid_lifetime', 'expire', 'subnet_id', 'fqdn_fwd', 'fqdn_rev',
                   'hostname', 'state')
        leases = {}
        try:
            leases_files = glob.glob(f"{self.path}/rootfs/etc/kea/dhcp4.leases*")
            for lease_file in leases_files:
                with open(lease_file, 'r') as f:
                    for raw_lease in f.read().splitlines()[1:]:
                        lease = dict(zip(headers, raw_lease.split(',')))
                        if lease['client_id']:
                            leases['client_id'] = lease
                        elif lease['hwaddr']:
                            leases['hwaddr'] = lease
        except (IOError, ValueError):
            pass
        return list(leases.values())

    def set_leases(self, leases: List[dict]):
        headers = 'address,hwaddr,client_id,valid_lifetime,expire,subnet_id,fqdn_fwd,fqdn_rev,hostname,state'
        file_to_set = f"{self.path}/rootfs/etc/kea/dhcp4.leases"
        with open(file_to_set, 'w') as f:
            f.write(headers)
            f.write('\n')
            for lease in leases:
                f.write(','.join(lease.values()))
                f.write('\n')
        return "Leases data has been set successfully.\n"

    def del_leases(self):
        leases_files = glob.glob(f"{self.path}/rootfs/etc/kea/dhcp4.leases*")
        if leases_files:
            for target_file in leases_files:
                try:
                    os.remove(target_file)
                except (OSError, Exception):
                    pass
        return "Leases data has been deleted successfully.\n"

    def get_networks(self) -> List['VNServiceNetwork']:
        """
        Получение списка интерфейсов контейнера
        """

        result = list()
        config = self._get_config()
        for idx in (0, 1):
            interface = VNServiceNetwork.get_configuration(config, idx)
            if interface:
                result.append(interface)
        return result

    def get_pid(self) -> Optional[str]:
        """
        Получние PID запущенного контейнера

        :return: вовращает PID запущенного контейнера или None, если контейнер не запущен
        """

        if not VNService.get_status(self.name) == VNService.STATUS.running:
            return None
        cmd = f"ps aux | grep {self.name}"
        raw_output = VNService._get_shell_output(cmd).splitlines()
        output = [line.split() for line in raw_output]
        for pid_info in output:
            if BASE_PATH in pid_info and self.name in pid_info:
                return pid_info[1]
        return None

    def set_config_item(self, param: str, value: str, save: bool = False) -> bool:
        """
        Установка значения параметра конфигурации контейнера

        :param param: название параметра
        :param value: значение параметра
        :param save: флаг указывает, надо ли сохранять конфигурацию после установки параметра
        :return: возвращает True, если параметр установлен, и False, если при установке значения параметра возникла
        ошибка
        """

        self.config[param] = value
        if save:
            config_str = [f"{k} = {v}" for k, v in self.config.items()]
            try:
                with open(self.config_file, 'w') as f:
                    f.write('\n'.join(config_str))
            except IOError:
                return False
        return True

    def save_config(self) -> bool:
        """
        Сохранение конфигурации контейнера в файл

        :return: возвращает True, если конфигурация сохранена, и False, если при сохранении конфигурации возникла ошибка
        """

        config_str = [f"{k} = {v}" for k, v in self.config.items()]
        try:
            with open(self.config_file, 'w') as f:
                f.write('\n'.join(config_str))
        except IOError:
            return False
        return True

    def get_config_item(self, param: str) -> Optional[str]:
        """
        Получение значения параметра конфигурации контейнера

        :param param: название параметра конфигурации
        :return: возвращает значение параметра конфигурации, если такой параметр существует, или None
        """

        if param == "lxc.rootfs.path":
            value = self.config[param][4:]
        else:
            value = self.config[param] if param in self.config.keys() else None
        return value

    def clear_config_item(self, param: str, save: bool = False) -> bool:
        """
        Удаление параметра из конфигурации контейнера

        :param param: название параметра конфигурации
        :param save: флаг указывает, надо ли сохранять конфигурацию после удаления параметра
        """

        if param in self.config.keys():
            self.config.pop(param)
        else:
            return False
        if save and not self.save_config():
            return False
        return True

    def clone(self, name: str) -> Optional['VNService']:
        """
        Клонирование базового шаблона контейнера

        :param name: название нового контейнера
        :return: возвращает клонированный контейнер в виде экземпляра класса VNService или None, если в процессе
        клонирования возникла ошибка
        """

        clone_path = f"{BASE_PATH}/{name}"
        cmd = f"mkdir {clone_path}"
        if not self._run_shell_cmd(cmd):
            return None
        cmd = f"cp -R {BASE_PATH}/vns-template/rootfs {clone_path}"
        pool = ThreadPool(processes=1)
        async_result = pool.apply_async(self._run_shell_cmd, (cmd,))
        clone_result = async_result.get(timeout=120)
        if not clone_result:
            cmd = f"rm -rf {clone_path}"
            self._run_shell_cmd(cmd)
            return None
        cmd = f"touch {clone_path}/config"
        if not self._run_shell_cmd(cmd):
            return None
        clone = VNService(name, initial=True)
        clone.set_config_item("lxc.rootfs.path", f"dir:{clone_path}/rootfs")
        clone.set_config_item("lxc.log.file", f"/var/log/veil/node/{name}.log")
        clone.set_config_item("lxc.uts.name", name)
        if not clone.save_config():
            clone.destroy()
            return None
        return clone

    def destroy(self) -> bool:
        """
        Удаление контейнера

        :return: возвращает True, если контейнер успешно удален, и False, если при удалении контейнера возникла ошибка
        """

        cmd = f"rm -rf {self.path}"
        if not self._run_shell_cmd(cmd):
            return False
        return True

    def _wait_status(self, status: str, timeout: int):
        """
        Ожидание статуса контейнера

        :param status: ожидаемый статус
        :param timeout:
        :return: возвращает True, если ожидаемый статус получен, и False, если при получении статуса возникла ошибка
        """

        cmd = f"lxc-wait -n {self.name} -s {status} -t {timeout}"
        if not self._run_shell_cmd(cmd):
            return False
        return True

    def _start(self) -> bool:
        """
        Запуск контейнера

        :return: возвращает True, если контейнер запущен, и False, если при запуске контейнера возникла ошибка
        """

        cmd = f"lxc-start -n {self.name}"
        if not self._run_shell_cmd(cmd):
            return False
        return True

    def _stop(self) -> bool:
        """
        Останов контейнера

        :return: возвращает True, если контейнер остановлен, и False, если при останове контейнера возникла ошибка
        """

        cmd = f"lxc-stop -n {self.name}"
        if not self._run_shell_cmd(cmd):
            return False
        return True

    def wait_started(self, timeout: int = 20) -> Optional[bool]:
        """
        Запуск контейнера с ожиданием статуса

        :param timeout: значение таймаута, по истечении которого задача считается не выполненной
        :return: возвращает True, если контейнер запущен, False, если до истечения таймаута контейнер не изменил статус,
        и None, если до истечения таймаута не был получен результат
        """

        executor = ThreadPoolExecutor(max_workers=2)
        wait_future = executor.submit(self._wait_status, VNService.STATUS.running, timeout)
        _ = executor.submit(self._start)
        try:
            result = wait_future.result(timeout)
            if result:
                self.status = VNService.STATUS.running
            return result
        except TimeoutError:
            return False

    def wait_stopped(self, timeout: int = 20) -> Optional[bool]:
        """
        Останов контейнера с ожиданием статуса

        :param timeout: значение таймаута, по истечении которого задача считается не выполненной
        :return: возвращает True, если контейнер остановлен, False, если до истечения таймаута контейнер не изменил
        статус, и None, если до истечения таймаута не был получен результат
        """

        executor = ThreadPoolExecutor(max_workers=2)
        wait_future = executor.submit(self._wait_status, VNService.STATUS.stopped, timeout)
        _ = executor.submit(self._stop)
        try:
            result = wait_future.result(timeout)
            if result:
                self.status = VNService.STATUS.running
                return result
            else:
                cmd = f"kill -9 {self.get_pid()}"
                return self._run_shell_cmd(cmd)
        except TimeoutError:
            cmd = f"kill -9 {self.get_pid()}"
            return self._run_shell_cmd(cmd)

    def attach_cmd(self, cmd: str, get_output: bool = False) -> Union[bool, str]:  # pylint: disable=W0621
        """
        Выполнение команды в контейнере

        :param cmd: команда, выполняемая в контейнере
        :param get_output: флаг, указывающий, надо ли получить вывод команды
        :return: возвращает True, если команда выполнена, и False, если при выполнении команды возникла ошибка; если
        передан флаг get_output, возвращает вывод команды
        """

        attach_cmd = f"lxc-attach -n {self.name} -- {cmd}"
        if get_output:
            ret = VNService._get_shell_output(attach_cmd)
            return ret
        else:
            if not self._run_shell_cmd(attach_cmd):
                return False
            return True

    @staticmethod
    def get_status(name) -> Optional[str]:
        """
        Получение текущего состояния контейнера

        :return: возвращает статус контейнера или None, в случае, если в сведениях о контейнере не содержится его
        текущего состояния
        """

        cmd = f"lxc-info -n {name}"
        ret = VNService._get_shell_output(cmd)
        raw_output = ret.splitlines()
        curr_status = next((row for row in raw_output if row.startswith("State")), None)
        if curr_status:
            state_row = curr_status.split(":")
            state = state_row[1].lstrip().rstrip()
            return state
        return None

    @classmethod
    def exists(cls, name: str) -> bool:
        """
        Проверка существования контейнера

        :param name: название контейнера
        :return: возвращает True, если контейнер существует, и False если не существует
        """

        cmd = "lxc-ls"
        output = cls._get_shell_output(cmd)
        if name in output.split():
            return True
        return False

    def create_interface(self, index: int, network_type: str, params: dict) -> bool:
        """
        Создание сетевого интерфейса контейнера

        :param index: индекс интерфейса
        :param network_type: тип интерфейса
        :param params: словарь, содержащий параметры интерфейса
        :return: возвращает True, если интерфейс создан, и False, если при создании интерфейса возникла ошибка
        """

        try:
            interface = VNServiceNetwork(index, network_type, **params)
            self.networks.append(interface)
            return True
        except (KeyError, Exception):
            return False

    def save_interface(self, index: int) -> bool:
        """
        Сохранение параметров интерфейса в конфигурацию контейнера

        :param index: индекс интерфейса
        :return: возвращает True, если сохранение успешно, и False, если при сохранении возникла ошибка
        """

        interface = next((net for net in self.networks if net.index == index), None)
        if not interface:
            return False
        self.set_config_item(f"lxc.net.{interface.index}.type", interface.network_type)
        self.set_config_item(f"lxc.net.{interface.index}.name", interface.name)
        self.set_config_item(f"lxc.net.{interface.index}.veth.pair", interface.pair)
        self.set_config_item(f"lxc.net.{interface.index}.flags", interface.flags)
        self.set_config_item(f"lxc.net.{interface.index}.hwaddr", interface.mac_address)
        self.set_config_item(f"lxc.net.{interface.index}.mtu", interface.mtu)
        self.set_config_item(f"lxc.net.{interface.index}.script.up", interface.script_up)
        self.set_config_item(f"lxc.net.{interface.index}.script.down", interface.script_down)
        if interface.ipv4_address:
            self.set_config_item(f"lxc.net.{interface.index}.ipv4.address", interface.ipv4_address)
            if interface.ipv4_gateway:
                self.set_config_item(f"lxc.net.{interface.index}.ipv4.gateway", interface.ipv4_gateway)
        if not self.save_config():
            return False
        return True

    def remove_interface(self, index: int) -> bool:
        """
        Удаление интерфейса контейнера

        :param index: индекс интерфейса
        :return: возвращает True, если удаление успешно, и False, если при удалении возникла ошибка
        """

        self.config = {k: v for k, v in self.config.items() if not k.startswith(f"lxc.net.{index}")}
        self.networks = self.get_networks()
        if not self.save_config():
            return False
        return True

    def repair_dhcp(self) -> None:
        """
        Восстановление работы сервиса dhcp посредством:
        - удаления файла /etc/kea/dhcp4.leases.pid, если сущетсвует
        """

        dhcp4_leases_pid_path = os.path.join(self.path, 'rootfs/etc/kea/dhcp4.leases.pid')
        if os.path.exists(dhcp4_leases_pid_path):
            os.remove(dhcp4_leases_pid_path)


class VNServiceNetwork:
    """
    Класс описывающий сеть (сетевой интерфейс) контейнера LXC
    """

    __slots__ = ('index', 'network_type', 'name', 'pair', 'flags', 'mac_address', 'mtu', 'script_up', 'script_down',
                 'ipv4_address', 'ipv4_gateway')

    def __init__(self, index: int, network_type: str, **kwargs):
        self.index = index
        self.network_type = network_type
        self.name = kwargs['name']
        self.pair = kwargs['pair']
        self.flags = kwargs['flags']
        self.mac_address = kwargs['mac_address']
        self.mtu = kwargs['mtu']
        self.script_up = kwargs['script_up']
        self.script_down = kwargs['script_down']
        self.ipv4_address = kwargs['ipv4_address']
        self.ipv4_gateway = kwargs['ipv4_gateway']

    @classmethod
    def get_configuration(cls, config: dict, index: int) -> Optional['VNServiceNetwork']:
        """
        Получение экземпляра класса из конфигурации контейнера
        :param config: конфигурация контейнера
        :param index: индекс интерфейса
        :return: возвращает экземпляр класса, если интерфейс есть в конфигурации контейнера или None
        """

        interface_config = {k: v for k, v in config.items() if k.startswith(f"lxc.net.{index}")}
        if interface_config:
            network_type = "veth"
            params = dict(
                name=interface_config[f"lxc.net.{index}.name"],
                pair=interface_config[f"lxc.net.{index}.veth.pair"],
                flags=interface_config[f"lxc.net.{index}.flags"],
                mac_address=interface_config[f"lxc.net.{index}.hwaddr"],
                mtu=interface_config[f"lxc.net.{index}.mtu"],
                script_up=interface_config[f"lxc.net.{index}.script.up"],
                script_down=interface_config[f"lxc.net.{index}.script.down"],
                ipv4_address=interface_config.get(f"lxc.net.{index}.ipv4.address", None),
                ipv4_gateway=interface_config.get(f"lxc.net.{index}.ipv4.gateway", None)
            )
            return VNServiceNetwork(index, network_type, **params)
        else:
            return None


class NetRuleSet(ABC):
    """
    Абстрактный класс, описывающий набор сетевых правил (блокирования трафика, зеркалирования портов и т.п.)
    """

    __slots__ = ('__config_dir',)

    def __init__(self):
        self.__config_dir = '/var/lib/ecp-veil/node/bin/networks'

    def config_dir(self):
        return self.__config_dir

    @abstractmethod
    def readable(self):
        pass

    @abstractmethod
    def find_rule(self, params: dict):
        pass

    @abstractmethod
    def remove_rule(self, rule, save: bool = False):
        pass

    @abstractmethod
    def add_rule(self, rule, save: bool = False):
        pass

    @abstractmethod
    def update_rule(self, rule, save: bool = False):
        pass

    @abstractmethod
    def save(self, apply: bool = False):
        pass

    @classmethod
    @abstractmethod
    def from_config_file(cls):
        pass

    @classmethod
    @abstractmethod
    def from_host_conf(cls):
        pass


class NetRule(ABC):
    """
    Абстрактный класс, описывающий сетевое правило (блокирования трафика, зеркалирования портов и т.п.)
    """

    __slots__ = ('applied', 'to_apply')

    def __init__(self, applied: bool = False, to_apply: bool = False):
        self.applied = applied
        self.to_apply = to_apply

    @property
    def slots(self):
        return set(chain.from_iterable(getattr(cls, "__slots__", ()) for cls in self.__class__.__mro__))

    @abstractmethod
    def readable(self):
        pass

    @abstractmethod
    def apply_rule(self):
        pass

    @abstractmethod
    def drop_rule(self):
        pass

    @abstractmethod
    def to_add_command(self):
        pass

    @abstractmethod
    def to_del_command(self):
        pass

    @classmethod
    @abstractmethod
    def from_json(cls, data: dict):
        pass


# region SriovStateRuleSet
class SriovStateRuleSet(NetRuleSet):
    """
    Класс, описывающий набор параметров для SR-IOV интерфейсов
    """

    __slots__ = ('__config_file', 'rules_set')

    def __init__(self):
        super(SriovStateRuleSet, self).__init__()
        self.__config_file = os.path.join(super().config_dir(), 'sriov-state.json')
        self.rules_set: List['SriovStateRule'] = []

    @property
    def config_file(self):
        return self.__config_file

    def readable(self) -> List[dict]:
        return [rule.readable() for rule in self.rules_set]

    def find_rule(self, params: dict) -> Optional['SriovStateRule']:
        interface = params.get('interface', None)
        if not interface:
            return None
        rule = next((rule for rule in self.rules_set if rule.interface == interface), None)
        return rule

    def remove_rule(self, rule: 'SriovStateRule', save: bool = False) -> bool:
        try:
            self.rules_set.remove(rule)
            if save and not self.save():
                return False
            return True
        except ValueError:
            return False

    def add_rule(self, rule: 'SriovStateRule', save: bool = False) -> bool:
        if not isinstance(rule, SriovStateRule):
            return False
        self.rules_set.append(rule)
        if save and not self.save():
            return False
        return True

    def update_rule(self, rule: 'SriovStateRule', save: bool = False) -> bool:
        index = None
        for idx, item in enumerate(self.rules_set):
            if item.interface == rule.interface:
                index = idx
        if index:
            self.rules_set[index] = rule
            return True
        return False

    def save(self, apply: bool = False) -> bool:
        try:
            if apply:
                for rule in filter(lambda x: x.to_apply and not x.applied, self.rules_set):
                    rule.apply_rule()
                    rule.applied = True
            with open(self.config_file, 'w') as f:
                f.write(json.dumps(self.readable()))
                f.write('\n')
            return True
        except OSError:
            return False

    @classmethod
    def from_config_file(cls) -> 'SriovStateRuleSet':
        rs = SriovStateRuleSet()
        try:
            with open(rs.config_file, 'r') as f:
                rules = json.loads(f.read())
        except json.decoder.JSONDecodeError:
            with open(rs.config_file, 'w') as f:
                f.write("[]\n")
            rules = []
        for rule in rules:
            rs.add_rule(SriovStateRule.from_json(rule))
        return rs

    @classmethod
    def from_host_conf(cls) -> 'SriovStateRuleSet':
        rs = SriovStateRuleSet()
        for device in (dev for dev in os.listdir(SYS_CLASS_NET_DIR)):
            if 'device' not in os.listdir(os.path.join(SYS_CLASS_NET_DIR, device)):
                continue
            dev_path = os.path.join(SYS_CLASS_NET_DIR, device, 'device')
            if 'sriov_totalvfs' in os.listdir(dev_path):
                cmd_ret = get_output(f'cat {os.path.join(dev_path, "sriov_numvfs")}')
                if not cmd_ret['result']:
                    continue
                if int(cmd_ret['comment']):
                    data = dict(interface=device, vfunctions=int(cmd_ret['comment']))
                    rs.add_rule(SriovStateRule.from_json(data))
        return rs


class SriovStateRule(NetRule):
    __slots__ = ('interface', 'vfunctions')

    def __init__(self, interface: str, vfunctions: int, applied: bool = False, to_apply: bool = False):
        super(SriovStateRule, self).__init__(applied, to_apply)
        self.interface = interface
        self.vfunctions = vfunctions

    def readable(self) -> dict:
        return {key: getattr(self, key) for key in self.slots}

    def apply_rule(self) -> bool:
        if not get_result(self.to_add_command()):
            return False
        return True

    def drop_rule(self) -> bool:
        if not get_result(self.to_del_command()):
            return False
        return True

    def to_add_command(self) -> str:
        return f"echo {self.vfunctions} > {os.path.join(SYS_CLASS_NET_DIR, self.interface, 'device', 'sriov_numvfs')}"

    def to_del_command(self) -> str:
        return f"echo 0 > {os.path.join(SYS_CLASS_NET_DIR, self.interface, 'device', 'sriov_numvfs')}"

    @classmethod
    def from_json(cls, data: dict) -> 'SriovStateRule':
        return SriovStateRule(**data)

# endregion


# region Port Mirroring
class PortMirroringRuleSet(NetRuleSet):
    __slots__ = ('__config_file', 'rules_set')

    def __init__(self):
        super(PortMirroringRuleSet, self).__init__()
        self.__config_file = os.path.join(super().config_dir(), 'port-mirrors.json')
        self.rules_set: List['PortMirroringRule'] = []

    @property
    def config_file(self):
        return self.__config_file

    def readable(self) -> List[dict]:
        return [rule.readable() for rule in self.rules_set]

    def find_rule(self, params: dict) -> Optional['PortMirroringRule']:
        source = params.get('source', None)
        target = params.get('target', None)
        direction = params.get('direction', None)
        pref_index = params.get('pref_index', None)
        if not source or not target or not direction or not pref_index:
            return None
        rule = next((rule for rule in self.rules_set if rule.source == source and rule.target == target and
                     rule.direction == direction and rule.pref_index == pref_index), None)
        return rule

    def remove_rule(self, rule: 'PortMirroringRule', save: bool = False) -> bool:
        try:
            self.rules_set.remove(rule)
            if save and not self.save():
                return False
            return True
        except ValueError:
            return False

    def add_rule(self, rule: 'PortMirroringRule', save: bool = False) -> bool:
        if not isinstance(rule, PortMirroringRule):
            return False
        self.rules_set.append(rule)
        if save and not self.save():
            return False
        return True

    def update_rule(self, rule: 'PortMirroringRule', save: bool = False) -> bool:
        index = None
        for idx, item in enumerate(self.rules_set):
            if item.source == rule.source:
                index = idx
        if index:
            self.rules_set[index] = rule
            return True
        return False

    def save(self, apply: bool = False) -> bool:
        try:
            if apply:
                for rule in filter(lambda x: x.to_apply and not x.applied, self.rules_set):
                    rule.apply_rule()
                    rule.applied = True
            with open(self.config_file, 'w') as f:
                f.write(json.dumps(self.readable()))
                f.write('\n')
            return True
        except OSError:
            return False

    @classmethod
    def from_config_file(cls) -> 'PortMirroringRuleSet':
        rs = PortMirroringRuleSet()
        try:
            with open(rs.config_file, 'r') as f:
                rules = json.loads(f.read())
        except json.decoder.JSONDecodeError:
            with open(rs.config_file, 'w') as f:
                f.write("[]\n")
            rules = []
        for rule in rules:
            rs.add_rule(PortMirroringRule.from_json(rule))
        return rs

    @classmethod
    def from_host_conf(cls) -> 'PortMirroringRuleSet':
        from veilrpc.states.networking import tc_mirred_rules_exists_three_directions
        rs = PortMirroringRuleSet()
        applied_rules = tc_mirred_rules_exists_three_directions()
        for direction, source_ports in applied_rules.items():
            for port, rules in source_ports.items():
                for rule in rules:
                    data = {'name': f"Port Mirror {len(rs.rules_set)}",
                            'source': port,
                            'pref_index': rule.get('pref_index'),
                            'direction': direction,
                            'target': rule.get('dst_port')
                            }
                    rs.add_rule(PortMirroringRule.from_json(data))
        return rs


class PortMirroringRule(NetRule):
    __slots__ = ('name', 'source', 'target', 'direction', 'pref_index')

    def __init__(self, name: str, source: str, target: str, direction: str, pref_index: dict,
                 applied: bool = False, to_apply: bool = False):
        super(PortMirroringRule, self).__init__(applied, to_apply)
        self.name = name
        self.source = source
        self.target = target
        self.direction = direction
        self.pref_index = pref_index

    def readable(self) -> dict:
        return {key: getattr(self, key) for key in self.slots}

    def apply_rule(self) -> bool:
        for cmd in self.to_add_command():
            if not get_result(cmd):
                return False
        return True

    def drop_rule(self) -> bool:
        for cmd in self.to_del_command():
            if not get_result(cmd):
                return False
        return True

    def to_add_command(self) -> List[str]:
        from veilrpc.states.networking import parse_tc_qdisc_exists
        qdisc_exists = states_result(parse_tc_qdisc_exists)
        if self.direction == 'ingress':
            cmd = [f"tc filter add dev {self.source} parent ffff: "
                   f"pref {self.pref_index['ingress']} protocol all u32 match u8 0 0 "
                   f"action mirred egress mirror dev {self.target}"]
            if self.source not in qdisc_exists['ingress']:
                cmd.insert(0, f"tc qdisc replace dev {self.source} ingress")
            return cmd
        elif self.direction == 'egress':
            cmd = [f"tc filter add dev {self.source} parent 1: "
                   f"pref {self.pref_index['egress']} protocol all u32 match u8 0 0 "
                   f"action mirred egress mirror dev {self.target}"]
            if self.source not in qdisc_exists['egress']:
                cmd.insert(0, f"tc qdisc replace dev {self.source} handle 1: root prio")
            return cmd
        elif self.direction == 'ingress-and-egress':
            cmd = [f"tc filter add dev {self.source} parent ffff: "
                   f"pref {self.pref_index['ingress']} protocol all u32 match u8 0 0 "
                   f"action mirred egress mirror dev {self.target}",
                   f"tc filter add dev {self.source} parent 1: pref {self.pref_index['egress']} "
                   f"protocol all u32 match u8 0 0 action mirred egress mirror dev {self.target}"]

            if self.source not in qdisc_exists['ingress']:
                cmd.insert(0, f"tc qdisc replace dev {self.source} ingress")
            if self.source not in qdisc_exists['egress']:
                cmd.insert(0, f"tc qdisc replace dev {self.source} handle 1: root prio")
            return cmd

    def to_del_command(self) -> List[str]:
        if self.direction in ('ingress', 'ingress-and-egress'):
            result = [f"tc filter delete dev {self.source} parent ffff: protocol all "
                      f"pref {self.pref_index['ingress']} u32"]
            if self.direction == 'ingress-and-egress':
                result.append(f"tc filter delete dev {self.source} parent 1: protocol all "
                              f"pref {self.pref_index['egress']} u32")
            return result
        else:
            return [f"tc filter delete dev {self.source} parent 1: protocol all pref {self.pref_index['egress']} u32"]

    @classmethod
    def from_json(cls, data: dict) -> 'PortMirroringRule':
        return PortMirroringRule(**data)

# endregion


# region VMachines QoS Policy Marking to vnetwork
class VMachinesQoSPolicyMarkingRuleSet(NetRuleSet):
    __slots__ = ('__config_file', 'rules_set')

    def __init__(self):
        super(VMachinesQoSPolicyMarkingRuleSet, self).__init__()
        self.__config_file = os.path.join(super().config_dir(), 'vmachines-qos-policy-marking.json')
        self.rules_set: List['VMachinesQoSPolicyMarkingRule'] = []

    @property
    def config_file(self):
        return self.__config_file

    def readable(self) -> List[dict]:
        return [rule.readable() for rule in self.rules_set]

    def find_rule(self, params: dict) -> Optional['VMachinesQoSPolicyMarkingRule']:
        # qos_policy = params.get('qos_policy', None)
        source = params.get('source', None)
        pref_index = params.get('pref_index', None)
        direction = params.get('direction', None)
        dscp = params.get('dscp', None)
        cos = params.get('cos', None)
        ethertype = params.get('ethertype', None)
        vlan_id = params.get('vlan_id', None)
        src_mac = params.get('src_mac', None)
        dst_mac = params.get('dst_mac', None)
        ip_protocol = params.get('ip_protocol', None)
        src_ip = params.get('src_ip', None)
        dst_ip = params.get('dst_ip', None)
        src_port = params.get('src_port', None)
        dst_port = params.get('dst_port', None)
        # is_enabled = params.get('is_enabled', None)
        # rule_index = params.get('rule_index', None)
        # id = params.get('id', None)
        # vnetwork = params.get('vnetwork', None)

        if not source or not pref_index or not direction or not ethertype:
            return None

        # rule = next((rule for rule in self.rules_set if rule.qos_policy == qos_policy and rule.source == source and
        #              rule.pref_index == pref_index and rule.direction == direction and rule.dscp == dscp and
        #              rule.cos == cos and rule.ethertype == ethertype and rule.vlan_id == vlan_id and
        #              rule.src_mac == src_mac and rule.dst_mac == dst_mac and rule.ip_protocol == ip_protocol and
        #              rule.src_ip == src_ip and rule.dst_ip == dst_ip and rule.src_port == src_port and
        #              rule.dst_port == dst_port and rule.is_enabled == is_enabled and rule.rule_index == rule_index),
        #             None)

        # Правило можно однозначано опеределить по параметрам source, pref_index, direction
        # Но будем сравнивать максимально возможные параметры
        rule = next((rule for rule in self.rules_set if rule.source == source and rule.pref_index == pref_index and
                     rule.direction == direction and rule.dscp == dscp and rule.cos == cos and
                     rule.ethertype == ethertype and rule.vlan_id == vlan_id and rule.src_mac == src_mac and
                     rule.dst_mac == dst_mac and rule.ip_protocol == ip_protocol and rule.src_ip == src_ip and
                     rule.dst_ip == dst_ip and rule.src_port == src_port and rule.dst_port == dst_port),
                    None)
        return rule

    def remove_rule(self, rule: 'VMachinesQoSPolicyMarkingRule', save: bool = False) -> bool:
        try:
            self.rules_set.remove(rule)
            if save and not self.save():
                return False
            return True
        except ValueError:
            return False

    def add_rule(self, rule: 'VMachinesQoSPolicyMarkingRule', save: bool = False) -> bool:
        if not isinstance(rule, VMachinesQoSPolicyMarkingRule):
            return False
        self.rules_set.append(rule)
        if save and not self.save():
            return False
        return True

    def update_rule(self, rule: 'VMachinesQoSPolicyMarkingRule', save: bool = False) -> bool:
        index = None
        for idx, item in enumerate(self.rules_set):
            # т.к. параметр pref_index уникален в нашей реализации, то т.о. правило можно однозначано опеределить по
            # этим параметрам ниже
            if item.source == rule.source and item.pref_index == rule.pref_index and item.direction == rule.direction:
                index = idx
        if index:
            self.rules_set[index] = rule
            return True
        return False

    def save(self, apply: bool = False) -> bool:
        try:
            if apply:
                for rule in filter(lambda x: x.to_apply and not x.applied, self.rules_set):
                    rule.apply_rule()
                    rule.applied = True
            with open(self.config_file, 'w') as f:
                f.write(json.dumps(self.readable()))
                f.write('\n')
            return True
        except OSError:
            return False

    @classmethod
    def from_config_file(cls) -> 'VMachinesQoSPolicyMarkingRuleSet':
        rs = VMachinesQoSPolicyMarkingRuleSet()
        try:
            with open(rs.config_file, 'r') as f:
                rules = json.loads(f.read())
        except json.decoder.JSONDecodeError:
            with open(rs.config_file, 'w') as f:
                f.write("[]\n")
            rules = []
        for rule in rules:
            rs.add_rule(VMachinesQoSPolicyMarkingRule.from_json(rule))
        return rs

    # Важный метод, здесь дополняем data реальными параметрами: 'source': port, 'direction': direction
    # и фэйковыми: 'qos_policy': "the_true_value_is_stored_in_json_file",
    # 'is_enabled': "the_true_value_is_stored_in_json_file", 'rule_index': "the_true_value_is_stored_in_json_file",
    # 'id': "the_true_value_is_stored_in_json_file"
    # Эти параметры добавляем, т.к. мы их не можем получить путем парсинга команды tc filter show
    # (функция tc_marking_rules_exists_params_needed)
    # Добавявление параметров делается для совместимости с классом VMachinesQoSPolicyMarkingRule
    # Полученные правила будут использоваться, в том числе, в методе find_rule
    @classmethod
    def from_host_conf(cls) -> 'VMachinesQoSPolicyMarkingRuleSet':
        from veilrpc.states.networking import tc_marking_rules_exists_params_needed
        rs = VMachinesQoSPolicyMarkingRuleSet()
        applied_rules = tc_marking_rules_exists_params_needed()
        for direction, source_ports in applied_rules.items():
            for port, rules in source_ports.items():
                for rule in rules:
                    data = {'qos_policy': "the_true_value_is_stored_in_json_file",
                            'source': port,
                            'pref_index': rule.get('pref_index'),
                            'direction': direction,
                            'dscp': rule.get('dscp'),
                            'cos': rule.get('cos'),
                            'ethertype': rule.get('ethertype'),
                            'vlan_id': rule.get('vlan_id'),
                            'src_mac': rule.get('src_mac'),
                            'dst_mac': rule.get('dst_mac'),
                            'ip_protocol': rule.get('ip_protocol'),
                            'src_ip': rule.get('src_ip'),
                            'dst_ip': rule.get('dst_ip'),
                            'src_port': rule.get('src_port'),
                            'dst_port': rule.get('dst_port'),
                            'is_enabled': "the_true_value_is_stored_in_json_file",
                            'rule_index': "the_true_value_is_stored_in_json_file",
                            'id': "the_true_value_is_stored_in_json_file",
                            'vnetwork': "the_true_value_is_stored_in_json_file"
                            }
                    rs.add_rule(VMachinesQoSPolicyMarkingRule.from_json(data))
        return rs


class VMachinesQoSPolicyMarkingRule(NetRule):
    # pylint: disable=redefined-builtin
    __slots__ = ('qos_policy', 'source', 'pref_index', 'direction', 'dscp', 'cos', 'ethertype', 'vlan_id', 'src_mac',
                 'dst_mac', 'ip_protocol', 'src_ip', 'dst_ip', 'src_port', 'dst_port', 'is_enabled', 'rule_index', 'id',
                 'vnetwork')

    def __init__(self, qos_policy: str, source: str, pref_index: int, direction: str, dscp: int, cos: int,
                 ethertype: str, vlan_id: int, src_mac: str, dst_mac: str, ip_protocol: str, src_ip: str, dst_ip: str,
                 src_port: int, dst_port: int, is_enabled: bool, rule_index: int, id: str, vnetwork: str,
                 applied: bool = False, to_apply: bool = False):
        super(VMachinesQoSPolicyMarkingRule, self).__init__(applied, to_apply)
        self.qos_policy = qos_policy
        self.source = source
        self.pref_index = pref_index
        self.direction = direction
        self.dscp = dscp
        self.cos = cos
        self.ethertype = ethertype
        self.vlan_id = vlan_id
        self.src_mac = src_mac
        self.dst_mac = dst_mac
        self.ip_protocol = ip_protocol
        self.src_ip = src_ip
        self.dst_ip = dst_ip
        self.src_port = src_port
        self.dst_port = dst_port
        self.is_enabled = is_enabled
        self.rule_index = rule_index
        self.id = id
        self.vnetwork = vnetwork

    def readable(self) -> dict:
        return {key: getattr(self, key) for key in self.slots}

    def apply_rule(self) -> bool:
        for cmd in self.to_add_command():
            if not get_result(cmd):
                return False
        return True

    def drop_rule(self) -> bool:
        for cmd in self.to_del_command():
            if not get_result(cmd):
                return False
        return True

    def to_add_command(self) -> List[str]:
        from veilrpc.states.networking import parse_tc_qdisc_exists
        qdisc_exists = states_result(parse_tc_qdisc_exists)

        rule_params = dict(pref_index=self.pref_index, direction=self.direction,
                           dscp=self.dscp, cos=self.cos, ethertype=self.ethertype, vlan_id=self.vlan_id,
                           src_mac=self.src_mac, dst_mac=self.dst_mac, ip_protocol=self.ip_protocol,
                           src_ip=self.src_ip, dst_ip=self.dst_ip, src_port=self.src_port, dst_port=self.dst_port)
        vminf_name = self.source
        cmd = [create_qos_marking_rule_to_valid_str(rule_params, vminf_name)]
        if self.direction == 'ingress':
            if self.source not in qdisc_exists['ingress']:
                cmd.insert(0, f"tc qdisc replace dev {self.source} ingress")
            return cmd
        elif self.direction == 'egress':
            if self.source not in qdisc_exists['egress']:
                cmd.insert(0, f"tc qdisc replace dev {self.source} handle 1: root prio")
            return cmd

    def to_del_command(self) -> List[str]:
        rule_params = dict(pref_index=self.pref_index, direction=self.direction, ethertype=self.ethertype)
        vminf_name = self.source
        cmd = [delete_qos_marking_rule_to_valid_str(rule_params, vminf_name)]
        return cmd

    @classmethod
    def from_json(cls, data: dict) -> 'VMachinesQoSPolicyMarkingRule':
        return VMachinesQoSPolicyMarkingRule(**data)

# endregion


# region VNetwork OvS QoS Policy
class VNetworkQoSPolicySet(NetRuleSet):
    __slots__ = ('__config_file', 'rules_set')

    def __init__(self):
        super(VNetworkQoSPolicySet, self).__init__()
        self.__config_file = os.path.join(super().config_dir(), VNETWORK_QOS_SHAPING_PARAMS)
        self.rules_set: List['VNetworkQoSPolicy'] = []

    @property
    def config_file(self) -> str:
        return self.__config_file

    def clone(self) -> 'VNetworkQoSPolicySet':
        rs = VNetworkQoSPolicySet()
        for r in self.rules_set:
            rs.add_rule(r)
        return rs

    def readable(self):
        return [rule.readable() for rule in self.rules_set]

    def find_rule(self, params: dict) -> Optional['VNetworkQoSPolicy']:
        qos = params.get('qos', None)
        port = params.get('port', None)
        if not port:
            return None
        if not qos:
            rule = next((rule for rule in self.rules_set if rule.port == port), None)
        else:
            rule = next((rule for rule in self.rules_set if rule.qos == qos and rule.port == port), None)
        return rule

    def add_rule(self, rule: 'VNetworkQoSPolicy', save: bool = False) -> bool:
        if not isinstance(rule, VNetworkQoSPolicy):
            return False
        if rule.readable() in self.readable():
            return True
        self.rules_set.append(rule)
        if save and not self.save():
            return False
        return True

    def remove_rule(self, rule: 'VNetworkQoSPolicy', save: bool = False) -> bool:
        if not isinstance(rule, VNetworkQoSPolicy):
            return False
        try:
            tmp_rs = self.clone()
            for r in tmp_rs.readable():
                if rule.readable() == r:
                    self.rules_set.remove(rule)
            if save and not self.save():
                return False
            return True
        except ValueError:
            return True

    def update_rule(self, rule: 'VNetworkQoSPolicy', save: bool = False) -> bool:
        if not isinstance(rule, VNetworkQoSPolicy):
            return False
        index = None
        for idx, item in enumerate(self.rules_set):
            if item.switch == rule.switch and item.port == rule.port:
                index = idx
        if index:
            self.rules_set[index] = rule
            return True
        return False

    def save(self, apply: bool = False) -> bool:
        result_rs = VNetworkQoSPolicySet()
        for rule in self.rules_set:
            if rule.readable() not in result_rs.readable():
                result_rs.add_rule(rule)
        try:
            if apply:
                for rule in filter(lambda x: x.to_apply and not x.applied, result_rs.rules_set):
                    rule.apply_rule()
                    rule.applied = True
            with open(self.config_file, 'w') as f:
                f.write(json.dumps(result_rs.readable()))
                f.write('\n')
            return True
        except OSError:
            return False

    @classmethod
    def from_config_file(cls) -> 'VNetworkQoSPolicySet':
        rs = VNetworkQoSPolicySet()
        try:
            with open(rs.config_file, 'r') as f:
                rules = json.load(f)
        except (json.decoder.JSONDecodeError, FileNotFoundError):
            with open(rs.config_file, 'w') as f:
                f.write("[]\n")
            rules = []
        for rule in rules:
            rs.add_rule(VNetworkQoSPolicy.from_json(rule))
        return rs

    @classmethod
    def from_host_conf(cls) -> Optional['VNetworkQoSPolicySet']:
        rs = VNetworkQoSPolicySet()
        existing_switches = get_vnetworks()
        for switch in existing_switches:
            existing_ports = get_ports_vm(switch)
            for port in existing_ports:
                qos = get_port_qos(port)
                rs.rules_set.append(VNetworkQoSPolicy(qos=qos['qos'], switch=switch, port=port,
                                                      rate=qos['rate'], to_apply=True))
        return rs


class VNetworkQoSPolicy(NetRule):
    # pylint: disable=redefined-builtin
    __slots__ = ('qos', 'switch', 'port', 'rate')

    def __init__(self, qos: str, switch: str, port: str, rate: int, applied: bool = False, to_apply: bool = False):
        super(VNetworkQoSPolicy, self).__init__(applied, to_apply)
        self.qos = qos
        self.switch = switch
        self.port = port
        self.rate = rate

    def readable(self) -> dict:
        return {key: getattr(self, key) for key in self.slots}

    def apply_rule(self) -> bool:
        for cmd in self.to_add_command():
            if not get_result(cmd):
                return False
        return True

    def drop_rule(self) -> bool:
        for cmd in self.to_del_command():
            if not get_result(cmd):
                return False
        return True

    def to_add_command(self) -> List[str]:
        return [
            f"ovs-vsctl set port {self.port} qos={self.qos}",
            f"ovs-vsctl set interface {self.port} ingress_policing_rate={int(self.rate/1000)}",
        ]

    def to_del_command(self) -> List[str]:
        return [
            f"ovs-vsctl clear port {self.port} qos",
            f"ovs-vsctl set interface {self.port} ingress_policing_rate=0",
        ]

    def qos_exists(self) -> bool:
        cmd = f"ovs-vsctl list qos {self.qos}"
        if not get_result(cmd):
            return False
        return True

    @classmethod
    def from_json(cls, data: dict) -> 'VNetworkQoSPolicy':
        return VNetworkQoSPolicy(**data)

# endregion


# region VNetwork ACL
@dataclass
class VNetworkOvSRuleSet:

    bridge: str
    base_flows: List[str] = field(default_factory=list)
    vm_acl_flows: List[str] = field(default_factory=list)
    vm_out_acl_flows: List[str] = field(default_factory=list)
    acl_flows: List[str] = field(default_factory=list)
    out_acl_flows: List[str] = field(default_factory=list)
    mark_flows: List[str] = field(default_factory=list)
    config: str = field(default_factory=str)
    config_vm: str = field(default_factory=str)

    def __post_init__(self):
        self.config = os.path.join(NET_PARAMS_DIR, f"{self.bridge}.json")
        if os.path.exists(self.config):
            self.init_from_config()
        else:
            self.init_from_ovs()
        self.config_vm = os.path.join(NET_PARAMS_DIR, f"{self.bridge}_vm.json")
        if os.path.exists(self.config_vm):
            self.init_from_config_vm()
        else:
            self.init_from_ovs_vm()

    @property
    def is_acl_mode(self) -> bool:
        cmd = f"ovs-ofctl --no-stats dump-flows {self.bridge} table={OvsTableNew.in_processing}"
        if (response := get_output(cmd=cmd, suppress_err=True)) and response['result'] and response['comment']:
            return True
        return False

    def init_from_ovs(self, names=False):
        fields = {'base_flows': [], 'acl_flows': [], 'out_acl_flows': [], 'mark_flows': []}
        names_flag = "--names" if names else ""
        if not self.is_acl_mode:
            # ACL mode off. Bridge at normal mode
            cmd = f"ovs-ofctl {names_flag} --no-stats dump-flows {self.bridge}"
            if (response := get_output(cmd, suppress_err=True)) and response['result'] and response['comment']:
                for r in response['comment'].splitlines():
                    fields['base_flows'].append(r.strip())
        else:
            # ACL mode on
            priority_pattern = re.compile(r"priority=(?P<priority>\d+)")
            table_pattern = re.compile(r"table=(?P<table>\d+)")
            cmd = f"ovs-ofctl {names_flag} --no-stats --rsort dump-flows {self.bridge}"
            if (response := get_output(cmd, suppress_err=True)) and response['result'] and response['comment']:
                for r in response['comment'].splitlines():
                    table = table_pattern.search(r)
                    priority = priority_pattern.search(r)
                    if table is None:
                        table = 0
                    else:
                        table = int(table.group('table'))
                    if priority is not None:
                        priority = int(priority.group('priority'))

                    if table is not None and priority is not None:
                        # ToDo: a.tsagan-manjiev: remove this hardcode (+4)
                        if (priority in range(OvsPriority.default, OvsPriority.default + 4) or
                                (table == OvsTableNew.in_acl and
                                 priority in range(min(OvsPriority.system_acl) & OvsPriority.pmask,
                                                   (max(OvsPriority.system_acl) & OvsPriority.pmask) + 1))):
                            fields['base_flows'].append(r.strip())

            # Extract not builtin input/output ACL flows
            for table in (*[OvsTableNew.in_acl + offset for offset in range(OVS_TABLE_MAX_OFFSET + 1)],
                          *[OvsTableNew.conntrack + offset for offset in range(OVS_TABLE_MAX_OFFSET + 1)],
                          OvsTableNew.out_acl):
                cmd = f"ovs-ofctl {names_flag} --no-stats --rsort dump-flows {self.bridge} table={table}"
                if (response := get_output(cmd, suppress_err=True)) and response['result'] and response['comment']:
                    for r in response['comment'].splitlines():
                        priority = priority_pattern.search(r)
                        if priority is not None:
                            priority = int(priority.groups()[0])
                            if not (priority in (OvsPriority.default, OvsPriority.default + 1) or (
                                    table == OvsTableNew.in_acl and
                                    priority in
                                    range(
                                        min(OvsPriority.system_acl) & OvsPriority.pmask,
                                        (max(OvsPriority.system_acl) & OvsPriority.pmask) + 1
                                    )
                            ) or (table == OvsTableNew.conntrack + OVS_TABLE_MAX_OFFSET and
                                  priority in
                                  range(
                                      min(OvsPriority.system_acl) & OvsPriority.pmask,
                                      (max(OvsPriority.system_acl) & OvsPriority.pmask) + 1
                                  ))):
                                if (table in
                                        [OvsTableNew.in_acl + offset for offset in range(OVS_TABLE_MAX_OFFSET + 1)] +
                                        [OvsTableNew.vm_in_acl + offset for offset in range(OVS_TABLE_MAX_OFFSET + 1)]):
                                    fields['acl_flows'].append(r.strip())
                                else:
                                    if 'in_port=' not in r and 'reg0=' not in r:
                                        fields['out_acl_flows'].append(r.strip())
        self.__dict__.update(fields)

    def init_from_ovs_vm(self):
        fields = {'vm_acl_flows': [], 'vm_out_acl_flows': []}
        if self.is_acl_mode:
            priority_pattern = re.compile(r"priority=(?P<priority>\d+)")
            # Extract not builtin input/output VM ACL flows
            for table in (*[OvsTableNew.vm_in_acl + offset for offset in range(OVS_TABLE_MAX_OFFSET + 1)],
                          *[OvsTableNew.vm_out_acl + offset for offset in range(OVS_TABLE_MAX_OFFSET + 1)],
                          OvsTableNew.out_processing):
                cmd = f"ovs-ofctl --names --no-stats --rsort dump-flows {self.bridge} table={table}"
                if (response := get_output(cmd, suppress_err=True)) and response['result'] and response['comment']:
                    for r in response['comment'].splitlines():
                        priority = priority_pattern.search(r)
                        if priority is not None:
                            priority = int(priority.groups()[0])
                            if priority not in (OvsPriority.default, OvsPriority.default + 1):
                                if (table in [
                                    OvsTableNew.vm_in_acl + offset for offset in range(OVS_TABLE_MAX_OFFSET + 1)] or
                                        'in_port=' in r):
                                    fields['vm_acl_flows'].append(r.strip().replace('"', ''))
                                else:
                                    fields['vm_out_acl_flows'].append(r.strip().replace('"', ''))
        self.__dict__.update(fields)

    def init_from_config(self) -> bool:
        try:
            with open(self.config, 'r') as f:
                fields = json.load(f)
        except (json.decoder.JSONDecodeError, FileNotFoundError):
            return False
        self.__dict__.update(fields)
        return True

    def init_from_config_vm(self) -> bool:
        try:
            with open(self.config_vm, 'r') as f:
                fields = json.load(f)
        except (json.decoder.JSONDecodeError, FileNotFoundError):
            return False
        self.__dict__.update(fields)
        return True

    def find_rule(self, search_flow: str) -> bool:
        fields = dataclasses.asdict(self)
        fields.pop('config')
        fields.pop('bridge')
        for _, flows in fields.items():
            for flow in flows:
                if "table=" not in flow:
                    flow = f"table=0,{flow}"
                if search_flow == flow.replace(", ", ",").replace(" ", ","):
                    return True
        return False

    def save(self) -> bool:
        fields = dataclasses.asdict(self)
        keys_to_pop = [key for key in fields.keys() if key[:3] == 'vm_'] + ['config', 'config_vm']
        for key in keys_to_pop:
            fields.pop(key)
        try:
            with open(self.config, 'w') as f:
                f.write(json.dumps(fields, indent=4))
                f.write('\n')
        except OSError:
            return False
        return True

    def save_vm(self) -> bool:
        fields = dataclasses.asdict(self)
        keys_to_pop = [key for key in fields.keys() if key[:3] != 'vm_' and key not in ['bridge']]
        for key in keys_to_pop:
            fields.pop(key)
        try:
            with open(self.config_vm, 'w') as f:
                f.write(json.dumps(fields, indent=4))
                f.write('\n')
        except OSError:
            return False
        return True

    def update_flows(self, flows: List[str], save: bool = False) -> bool:
        """
        flows - список OpenFlow правил с delete_strict/add для применения к таблице OpenFlow
        """

        update_file = os.path.join(NET_PARAMS_DIR, f"{self.bridge}.tmp")
        with open(update_file, 'w') as f:
            f.write('\n'.join(flows))
        cmd = f"ovs-ofctl add-flows {self.bridge} {update_file}"
        if not get_result(cmd):
            return False
        os.remove(update_file)
        if save:
            self.init_from_ovs()
            self.save()
        return True

    def update_flows_vm(self, flows: List[str], skip_list: List[bool] = None, save: bool = False):
        """
        flows - список OpenFlow правил для ВМ с delete_strict/add для применения к таблице OpenFlow
        """

        # Список правил, которые можно применить на ноде (нода с запущенной VM)
        flows_to_apply = []
        # Список правил, которые нельзя применить на ноде, но необходимо сохранить (нода с незапущенной VM)
        flows_to_save = []
        out_flows_to_save = []
        if skip_list and len(skip_list) == len(flows):
            pattern = re.compile(r"(?P<table0>table=\d+,)\S|(?P<table1>table=\d+,)(?P<spaces>\s+)")
            for i, flow in enumerate(flows):
                # Если правило можно применить на ноде (нода с запущенной VM)
                if not skip_list[i]:
                    if "reg0=" in flow:
                        """
                        Если правило содержит информацию об исходящем интерфейсе,
                        то конвертируем имя интерфейса в цифру
                        """
                        flow = convert_vm_out_flow(flow)
                    flows_to_apply.append(flow)
                else:
                    """
                    Если правило нельзя применить на ноде (нода с незапущенной VM)
                    и это правило на добавление ACL
                    """
                    if "delete_strict " not in flow:
                        match = pattern.search(flow)
                        flow = flow.replace(",actions=", " actions=").replace("add ", "")
                        if match and match.group('table0'):
                            if "reg0=" in flow:
                                out_flows_to_save.append(flow.replace(match.group('table0'),
                                                                      f"{match.group('table0')} "))
                            else:
                                flows_to_save.append(flow.replace(match.group('table0'), f"{match.group('table0')} "))
                        if match and match.group('table1'):
                            if "reg0=" in flow:
                                out_flows_to_save.append(flow.replace(match.group('spaces'), " "))
                            else:
                                flows_to_save.append(flow.replace(match.group('spaces'), " "))
        # Сохраняем набор правило для применения во временный файл и применяем в OpenFlow
        update_file = os.path.join(NET_PARAMS_DIR, f"{self.bridge}_vm.tmp")
        with open(update_file, 'w') as f:
            f.write('\n'.join(flows_to_apply))
        cmd = f"ovs-ofctl add-flows {self.bridge} {update_file}"
        if not get_result(cmd):
            return False
        os.remove(update_file)

        if save:
            self.init_from_ovs_vm()
            self.vm_acl_flows.extend(flows_to_save)
            pattern = re.compile(r"reg0=(?P<port_number>0x[a-z0-9]+)", re.IGNORECASE)
            for flow in self.vm_out_acl_flows:
                if match := pattern.search(flow):
                    # Перед сохранением OUT VM ACL в файл конвертируем цифру интерфейса в его имя в OvS
                    vm_interface_name = get_interface_name_by_number(self.bridge, int(match.group('port_number'), 16))
                    if vm_interface_name:
                        out_flows_to_save.append(flow.replace(f"reg0={match.group('port_number')}",
                                                              f"reg0={vm_interface_name}"))
            self.vm_out_acl_flows = out_flows_to_save
            self.save_vm()
        return True

    def add_flow(self, flow: str, save: bool = False) -> bool:
        cmd = f"ovs-ofctl add-flow {self.bridge} '{flow}'"
        if not get_result(cmd):
            return False
        if save:
            self.init_from_ovs()
            self.save()
        return True

    def delete_flow(self, flow: str, save: bool = False) -> bool:
        cmd = f"ovs-ofctl --strict del-flows {self.bridge} '{flow.split('actions')[0]}'"
        if not get_result(cmd):
            return False
        if save:
            self.init_from_ovs()
            self.save()
        return True

    def erase(self) -> bool:
        cmd = f"ovs-ofctl del-flows {self.bridge}"
        if not get_result(cmd):
            return False
        return True

    def reset(self, save: bool = False) -> bool:
        self.save()
        self.save_vm()
        if not self.erase():
            VNetworkOvSRuleSet.sync_with_config(bridge=self.bridge)
            return False
        if not self.add_flow(OVS_FLOW_PRESETS_NEW['DEFAULT']):
            VNetworkOvSRuleSet.sync_with_config(bridge=self.bridge)
            return False
        if save:
            self.save()
            self.save_vm()
        return True

    def acl_mode(self, mode: bool = True, save: bool = False) -> bool:
        if self.is_acl_mode == mode:
            return True
        self.save()
        self.save_vm()
        # Удалим правила или откатимся к конфигу в случае отказа
        if not self.erase():
            VNetworkOvSRuleSet.sync_with_config(bridge=self.bridge)
            return False
        # Накатим правила или откатимся к конфигу в случае отказа
        if mode:
            for flow in OVS_FLOW_PRESETS_NEW['ACL_MODE']:
                if not self.add_flow(flow):
                    VNetworkOvSRuleSet.sync_with_config(bridge=self.bridge)
                    return False
        else:
            if not self.reset():
                return False
        # Всё хорошо, сохраняемся.
        if save:
            self.init_from_ovs()
            self.init_from_ovs_vm()
            self.save()
            self.save_vm()
        return True

    def acl_options(self, options: dict) -> bool:
        mode = options.get('mode', False)
        if self.is_acl_mode != mode:
            if not self.acl_mode(mode, save=True):
                return False
        if mode:
            flows = []
            if (arp := options.get('arp')) is not None:
                if arp:
                    flows.append(f"add {OVS_FLOW_PRESETS_NEW['ARP_ALLOW']}")
                else:
                    flows.append(f"delete_strict {OVS_FLOW_PRESETS_NEW['ARP_ALLOW'].split('actions')[0]}")
            if (mac_learning := options.get('mac_learning')) is not None:
                if mac_learning:
                    for flow in OVS_FLOW_PRESETS_NEW['MAC_LEARNING_ENABLE']:
                        flows.append(f"add {flow}")
                else:
                    for flow in OVS_FLOW_PRESETS_NEW['MAC_LEARNING_DISABLE']:
                        flows.append(f"add {flow}")
            if (stp_drop := options.get('stp_drop')) is not None:
                if stp_drop:
                    flows.append(f"add {OVS_FLOW_PRESETS_NEW['STP_DROP']}")
                else:
                    flows.append(f"delete_strict {OVS_FLOW_PRESETS_NEW['STP_DROP'].split('actions')[0]}")
            if (broadcast_drop := options.get('broadcast_drop')) is not None:
                if broadcast_drop:
                    flows.append(f"add {OVS_FLOW_PRESETS_NEW['BROADCAST_SRC_DROP']}")
                else:
                    flows.append(f"delete_strict {OVS_FLOW_PRESETS_NEW['BROADCAST_SRC_DROP'].split('actions')[0]}")
            if (conntrack := options.get('conntrack')) is not None:
                if conntrack:
                    for flow in OVS_FLOW_PRESETS_NEW['CONNTRACK_DISABLE']:
                        flows.append(f"delete_strict {flow.split('actions')[0]}")
                    for flow in OVS_FLOW_PRESETS_NEW['CONNTRACK_ENABLE']:
                        flows.append(f"add {flow.format(VNET_MARK=self.bridge.replace('vnet_', ''))}")
                else:
                    for flow in OVS_FLOW_PRESETS_NEW['CONNTRACK_ENABLE']:
                        flows.append(f"""delete_strict {flow.format(
                            VNET_MARK=self.bridge.replace('vnet_', '')).split('actions')[0]}""")
                    for flow in OVS_FLOW_PRESETS_NEW['CONNTRACK_DISABLE']:
                        flows.append(f"add {flow}")
            if flows:
                return self.update_flows(flows, save=True)
        else:
            return self.reset(save=True)
        return False

    def dump_flows(self, table: int = OvsTableNew.in_acl, base_flows: bool = False) -> List[str]:
        if (table in [OvsTableNew.in_acl + offset for offset in range(OVS_TABLE_MAX_OFFSET + 1)] +
                [OvsTableNew.out_processing]):
            if not base_flows:
                dump = list(map(lambda f: f.strip().replace(' actions', ',actions').replace(' ', ''), self.acl_flows))
            else:
                dump = []
                for offset in range(OVS_TABLE_MAX_OFFSET + 1):
                    dump += list(map(lambda f: f.strip().replace(' actions', ',actions').replace(' ', ''),
                                     filter(lambda f: f"table={table + offset}," in f, self.base_flows)))
        elif (table in [OvsTableNew.vm_in_acl + offset for offset in range(OVS_TABLE_MAX_OFFSET + 1)] +
              [OvsTableNew.out_processing]):
            if not base_flows:
                dump = list(map(lambda f: f.strip().replace(' actions', ',actions').replace(' ', ''),
                                self.vm_acl_flows))
            else:
                dump = []
                for offset in range(OVS_TABLE_MAX_OFFSET + 1):
                    dump += list(map(lambda f: f.strip().replace(' actions', ',actions').replace(' ', ''),
                                     filter(lambda f: f"table={table + offset}," in f, self.base_flows)))
        elif (table in [OvsTableNew.vm_out_acl + offset for offset in range(OVS_TABLE_MAX_OFFSET + 1)] +
              [OvsTableNew.out_processing]):
            if not base_flows:
                dump = list(map(lambda f: f.strip().replace(' actions', ',actions').replace(' ', ''),
                                self.vm_out_acl_flows))
            else:
                dump = []
                for offset in range(OVS_TABLE_MAX_OFFSET + 1):
                    dump += list(map(lambda f: f.strip().replace(' actions', ',actions').replace(' ', ''),
                                     filter(lambda f: f"table={table + offset}," in f, self.base_flows)))
        elif (table in [OvsTableNew.conntrack + offset for offset in range(OVS_TABLE_MAX_OFFSET + 1)] +
              [OvsTableNew.out_acl]):
            if not base_flows:
                dump = list(map(lambda f: f.strip().replace(' actions', ',actions').replace(' ', ''),
                                self.out_acl_flows))
            else:
                dump = list(map(lambda f: f.strip().replace(' actions', ',actions').replace(' ', ''),
                                filter(lambda f: f"table={table}," in f, self.base_flows)))
        else:
            dump = list(map(lambda f: f.strip().replace(' actions', ',actions').replace(' ', ''),
                            filter(lambda f: f"table={table}," in f, self.base_flows)))
        return dump

    @classmethod
    def sync_with_config(cls, bridge: str) -> bool:
        data_ovs = VNetworkOvSRuleSet(bridge=bridge)
        data_ovs.init_from_ovs()
        data_ovs.init_from_ovs_vm()

        # Конвертируем номер интерфейса в его имя в OvS
        pattern = re.compile(r"reg0=(?P<port_number>0x[a-z0-9]+)", re.IGNORECASE)
        out_flows = []
        for flow in data_ovs.vm_out_acl_flows:
            if match := pattern.search(flow):
                vm_interface_name = get_interface_name_by_number(data_ovs.bridge, int(match.group('port_number'), 16))
                if vm_interface_name:
                    out_flows.append(flow.replace(f"reg0={match.group('port_number')}", f"reg0={vm_interface_name}"))
        data_ovs.vm_out_acl_flows = out_flows

        data_config = VNetworkOvSRuleSet(bridge=bridge)
        if not data_config.init_from_config() and not data_config.init_from_config_vm():
            # Не удалось получить конфиг
            return False

        if DeepDiff(data_ovs.__dict__, data_config.__dict__, ignore_order=True) == {}:
            # Различий нет
            return True
        else:
            # Готовим правила для удаления
            flows_to_delete = {
                'base_flows': list(map(
                    lambda f: f"delete_strict {f.split('actions')[0]}",
                    set(data_ovs.base_flows) - set(data_config.base_flows))),
                'acl_flows': list(map(
                    lambda f: f"delete_strict {f.split('actions')[0]}",
                    set(data_ovs.acl_flows) - set(data_config.acl_flows))),
                'out_acl_flows': list(map(
                    lambda f: f"delete_strict {f.split('actions')[0]}",
                    set(data_ovs.out_acl_flows) - set(data_config.out_acl_flows))),
                'mark_flows': list(map(
                    lambda f: f"delete_strict {f.split('actions')[0]}",
                    set(data_ovs.mark_flows) - set(data_config.mark_flows)))
            }
            flows_to_delete_vm = {
                'vm_acl_flows': list(map(
                    lambda f: f"delete_strict {f.split('actions')[0]}",
                    diff_flows(data_ovs.vm_acl_flows, data_config.vm_acl_flows))),
                'vm_out_acl_flows': []
            }
            data_config_vm_out_acl_flows = []
            for i in range(len(data_config.vm_out_acl_flows)):
                data_config_vm_out_acl_flows.append(convert_vm_out_flow(data_config.vm_out_acl_flows[i]))
            flows_to_delete_vm['vm_out_acl_flows'] = list(map(
                lambda f: f"delete_strict {f.split('actions')[0]}",
                diff_flows(data_ovs.vm_out_acl_flows, data_config_vm_out_acl_flows)))

            # Готовим правила для добавления
            flows_to_add = {
                'base_flows': list(map(lambda f: f"add {f}", set(data_config.base_flows) - set(data_ovs.base_flows))),
                'acl_flows': list(map(lambda f: f"add {f}", set(data_config.acl_flows) - set(data_ovs.acl_flows))),
                'out_acl_flows': list(map(
                    lambda f: f"add {f}",
                    set(data_config.out_acl_flows) - set(data_ovs.out_acl_flows))),
                'mark_flows': list(map(lambda f: f"add {f}", set(data_config.mark_flows) - set(data_ovs.mark_flows)))
            }
            flows_to_add_vm = {
                'vm_acl_flows': list(map(
                    lambda f: f"add {f}",
                    diff_flows(data_config.vm_acl_flows, data_ovs.vm_acl_flows))),
                'vm_out_acl_flows': []
            }
            for i in range(len(data_config.vm_out_acl_flows)):
                if not flow_in_set(data_config_vm_out_acl_flows[i], data_ovs.vm_out_acl_flows):
                    flows_to_add_vm['vm_out_acl_flows'].append(f"add {data_config.vm_out_acl_flows[i]}")

            # Собираем всё в один список
            flows = []
            for _, v in flows_to_add.items():
                flows.extend(v)
            for _, v in flows_to_delete.items():
                flows.extend(v)

            valid_flows_vm = []
            skip_list = []
            for _, v in flows_to_delete_vm.items():
                valid_flows_vm.extend(v)
                for _ in range(len(v)):
                    skip_list.append(False)
            for _, v in flows_to_add_vm.items():
                for flow in v:
                    skip = False
                    response = states_result(get_output, cmd=f'ovs-vsctl list-ports {bridge}').splitlines()
                    match = OVF_RULE_PATTERN_VM_PORT.search(flow)
                    port = val if (val := match.group('in_port')) else match.group('out_port')
                    if port not in response:
                        skip = True
                    valid_flows_vm.append(flow)
                    skip_list.append(skip)

            # Обновляем
            if not data_ovs.update_flows(flows):
                return False
            if not data_ovs.update_flows_vm(flows=valid_flows_vm, skip_list=skip_list):
                return False
        return True

# endregion


# region OvS Mirroring
class OvSMirror:
    """
    Класс для управления настройками зеркалирования через OvS.
    """
    def __init__(self, name: str, source_interface: str, destination_interface: str, direction: str = 'both'):
        """
        Initialize the class with the specified source interface and optional destination interface and direction.

        Parameters:
            source_interface (str): The source interface for the mirroring.
            destination_interface (str): The destination interface for the mirroring.
            direction (str, optional): The direction of the packets. Defaults to 'both'. Possible: 'both', 'in', 'out'.
        """

        self.name = name
        self.source_interface = source_interface
        self.destination_interface = destination_interface
        self.direction = direction

    def __get_bridge_by_interface(self, interface: str = None) -> str:
        """
        Возвращает имя бриджа по интерфейсу.
        Parameters:
            interface (str, optional): The interface name. Defaults to None.
        """
        if not interface:
            interface = self.source_interface
        cmd = f"ovs-vsctl iface-to-br {interface}"
        output = get_output(cmd)
        if not output['result']:
            raise CommandExecutionError(output['comment'].strip())
        return output['comment'].strip()

    @staticmethod
    def __instance_exists(instance: str, name: str) -> bool:
        """
        Проверяет существование интерфейса.
        Parameters:
            instance (str): The instance (OvS table name).
            name (str): The instance name.
        """
        cmd = f"ovs-vsctl --if-exists get {instance} {name} name"
        output = get_output(cmd)
        return output['comment'].strip() == name

    def create(self) -> bool:
        """
        Создает зеркальный бридж.
        """

        # Проверяем существование интерфейсов, бриджа и зеркалирования. Если нет интерфейса назначения, то создаем.
        if self.__instance_exists('Mirror', self.name):
            raise CommandExecutionError(f'Mirror {self.name} already exists!')
        if not self.__instance_exists('Interface', self.source_interface):
            raise CommandExecutionError(f'Interface {self.source_interface} not found!')
        bridge_by_source = self.__get_bridge_by_interface()
        if not self.__instance_exists('Bridge', bridge_by_source):
            raise CommandExecutionError(f'Bridge {bridge_by_source} not found!')
        if not self.__instance_exists('Interface', self.destination_interface):
            cmd = (f"ovs-vsctl add-port {bridge_by_source} {self.destination_interface} "
                   f"-- set Interface {self.destination_interface} type=internal")
            if not get_result(cmd):
                raise CommandExecutionError(f'Failed to create destination interface {self.destination_interface}!')
        else:
            bridge_by_destination = self.__get_bridge_by_interface(self.destination_interface)
            if not self.__instance_exists('Bridge', bridge_by_destination):
                raise CommandExecutionError(f'Bridge {bridge_by_destination} not found!')
            if bridge_by_source != bridge_by_destination:
                raise CommandExecutionError(f'Bridge {bridge_by_source} and {bridge_by_destination} are not the same!')

        # Создаем правило зеркалирования
        mirror_cmd = (f"ovs-vsctl"
                      f" -- --id=@m0 create Mirror name={self.name}"
                      f" -- add Bridge {bridge_by_source} mirrors @m0"
                      f" -- --id=@dst_port get Port {self.destination_interface}"
                      f" -- --id=@src_port get Port {self.source_interface}"
                      f" -- set mirror {self.name} select_src_port=@src_port select_dst_port=@src_port"
                      f" output_port=@dst_port")
        mirror_create = get_result(mirror_cmd)
        destination_interface_cmd = f"ip link set {self.destination_interface} up"
        destination_interface_up = get_result(destination_interface_cmd)
        return mirror_create and destination_interface_up

    def remove(self) -> bool:
        # Удаляем правило зеркалирования
        mirror_remove = destination_remove = True
        if self.__instance_exists('Mirror', self.name):
            mirror_remove_cmd = (
                f"ovs-vsctl"
                f" -- --id=@m0 get Mirror {self.name}"
                f" -- remove Bridge {self.__get_bridge_by_interface()} mirrors @m0")
            mirror_remove = get_result(mirror_remove_cmd)
        if self.__instance_exists('Interface', self.destination_interface):
            destination_interface_cmd = f"ip link set {self.destination_interface} down"
            get_result(destination_interface_cmd)
            destination_remove_cmd = (
                f"ovs-vsctl"
                f" --if-exists del-port {self.__get_bridge_by_interface()} {self.destination_interface}")
            destination_remove = get_result(destination_remove_cmd)

        return mirror_remove and destination_remove

# endregion
